import { Instruction, Register, Value } from "./ir";
import { assertExhaustive } from "./utils";

export class X8664Generator {
    private result = "";

    public generate(instructions: Instruction[]): string {
        this.result = "";
        for (const ins of instructions) {
            this.generateInstruction(ins, instructions);
        }
        return this.asmWithHeaders(this.result);
    }

    private generateInstruction(ins: Instruction, instructions: Instruction[]) {
        let result = "";
        if (ins.jumpedTo)
            result += `.L${ins.label}:\n`;
        switch (ins.type) {
            case "load":
                this.add(`
                    ; load
                    mov ${this.reg64(ins.register)}, ${this.value(ins.value)}
                `);
                break;
            case "push":
                this.add(`
                    ; push
                    push ${this.reg64(ins.register)}
                `);
            case "pop":
                this.add(`
                    ; pop
                    pop ${this.reg64(ins.register)}
                `);
                break;
            case "negate":
                this.add(`
                    ; neg
                    mov ${this.reg64(ins.dest)}, ${this.reg64(ins.src)}
                    neg ${this.reg64(ins.dest)}
                `);
                break;
            case "add":
                this.add(`
                    ; add
                    mov ${this.reg64(ins.dest)}, ${this.reg64(ins.left)}
                    add ${this.reg64(ins.dest)}, ${this.reg64(ins.right)}
                `);
                break;
            case "sub":
                this.add(`
                    ; sub
                    mov ${this.reg64(ins.dest)}, ${this.reg64(ins.left)}
                    sub ${this.reg64(ins.dest)}, ${this.reg64(ins.right)}
                `);
                break;
            case "mul":
                this.add(`
                    ; mul
                    mov ${this.reg64(ins.dest)}, ${this.reg64(ins.left)}
                    imul ${this.reg64(ins.dest)}, ${this.reg64(ins.right)}
                `);
                break;
            case "div":
                this.add(`
                    ; div
                    mov rdi, ${this.reg64(ins.right)}
                    mov rax, ${this.reg64(ins.left)}
                    xor rdx, rdx
                    cqo
                    idiv rdi
                    mov ${this.reg64(ins.dest)}, rax
                `);
                break;
            case "jump_zero":
                this.add(`
                    ; jump_zero
                    cmp ${ins.register}
                    jz .L${(() => {
                        const dest = instructions[ins.location];
                        if (!dest.jumpedTo)
                            throw new Error("impossible");
                        return dest.label;
                    })()}
                `);
                break;
            case "jump":
                this.add(`
                    ; jump
                    jz .L${(() => {
                        const dest = instructions[ins.location];
                        if (!dest.jumpedTo)
                            throw new Error("impossible");
                        return dest.label;
                    })()}
                `);
                break;
            default:
                assertExhaustive(ins);
        }
        return result;
    }

    private add(lines: string) {
        this.result += lines
            .trim()
            .split("\n")
            .map(s => s.replace(/^(    )+/, "    "))
            .join("\n")
            + "\n";
    }

    private asmWithHeaders(asm: string) {
        return `
            bits 64
            global _start

            _start:
            \t${asm}
            exit:
            \tmov rdi, rax
            \tmov rax, 60
            \tsyscall

        `.replace(/ /g, "").replace(/\t/g, "    ");
    }

    private reg64(reg: Register): string {
        switch (reg) {
            case "acc": return "rax";
            case "op": return "rdx";
        }
    }

    private reg32(reg: Register): string {
        switch (reg) {
            case "acc": return "eax";
            case "op": return "edx";
        }
    }


    private value(value: Value): string {
        return value.toString();
    }

}
